%% 1d linear elastic bar element
%
% Numerical Analysis of Structures WS 2017
%
% (c) 2017 Marek Tyburec, marek.tyburec@fsv.cvut.cz

function maxError = elasticBar1(numberOfElements, plotting)

%% DEFINE STRUCTURE
% nodal locations in global coordinates [m]
x0 = 0; xE = 1;
x = linspace(x0, xE, numberOfElements+1);

% nodal (external) forces in global coordinates [N]
nodeNumberFx = [];
Fx = [];

% intensity of distributed loading in nodes [N]
nodeNumberfx = 1:numel(x);
fx = 1*ones(1, numel(x));

% prescribed nodal displacements in global coordinates [m]
nodeNumberUx = 1;
Ux = 0;

% elements (node numbers)
elements = horzcat((1:numel(x)-1)', (2:numel(x))');

% nodal cross section of elements [m2]
A = ones(numberOfElements, 1);

% young modulus of bars [Pa]
E = ones(numberOfElements,1);

% auxiliary variables
numberOfNodes = length(x);

%% FEM ANALYSIS
% prealocation
% number of degrees of freedom per node=1 (horizontal displacement)
stiffnessMatrixGlobal = zeros(1*numberOfNodes, 1*numberOfNodes);
elementLengths = zeros(numberOfElements,1);
forceVector = zeros(numberOfNodes,1);

% assembly of the global stiffness matrix
% for all elements
for element=1:numberOfElements
  % pick the element 'element'
  
  % find starting and ending node of the element
  FirstNode = elements(element, 1);
  SecondNode = elements(element, 2);
  
  % code numbers of the node: [FirstNode SecondNode]
  codeNumbersElement = [FirstNode SecondNode];
                        
  % length of the element
  elementLengths(element) = x(SecondNode) - x(FirstNode);
    
  % element stiffness matrix in global coordinates
  stiffnessMatrixElement = E(element) * A(element) / elementLengths(element) * ...
                                                                [ 1 -1; ...
                                                                 -1  1];
     
  % element body force vector
  bodyForceVectorElement = elementLengths(element)/6 * [2 1; 1 2] * [fx(codeNumbersElement(1)); ...
                                                                     fx(codeNumbersElement(2))];
                            
  % assembly (add element stiffness matrix to the structural global stiffness matrix)
  stiffnessMatrixGlobal(codeNumbersElement, codeNumbersElement) = ...
      stiffnessMatrixGlobal(codeNumbersElement, codeNumbersElement) + stiffnessMatrixElement;
      
  % assembly of force vector
  forceVector(codeNumbersElement) = forceVector(codeNumbersElement) + bodyForceVectorElement;
  
end

% add nodal forces
forceVector(nodeNumberFx) = Fx;

% prescribed displacement vector
displacementVector = zeros(numberOfNodes,1);
displacementVector(nodeNumberUx) = Ux;

% essential code numbers (prescribed displacement)
essentialBoundaryConditions = nodeNumberUx;

% free boundary conditions
freeBoundaryConditions = setdiff(1:numberOfNodes, essentialBoundaryConditions);

% solution
displacementVector(freeBoundaryConditions) = stiffnessMatrixGlobal(freeBoundaryConditions, freeBoundaryConditions) \ ... 
                                          (forceVector(freeBoundaryConditions) - ...
                                          stiffnessMatrixGlobal(freeBoundaryConditions,essentialBoundaryConditions)*...
                                          displacementVector(essentialBoundaryConditions));

reactions = stiffnessMatrixGlobal(essentialBoundaryConditions, essentialBoundaryConditions) * ...
            displacementVector(essentialBoundaryConditions) + ...
            stiffnessMatrixGlobal(essentialBoundaryConditions,freeBoundaryConditions) * displacementVector(freeBoundaryConditions);
                        
%% POST-PROCESSING
strains = (displacementVector(elements(:,2)) - displacementVector(elements(:,1)))./elementLengths;
stresses = E .* strains;

% plot figure with exact displacements
xExact = linspace(0,1,1e5);
uExact = -1*xExact.^2/2 + xExact;
if plotting
    hold on;
    plot(xExact, uExact, '--k');
    plot(x, displacementVector, '-r');
    hold off;
end

% plot figure with exact stress
xExact = linspace(0,1,1e5);
sigmaExact = -xExact + 1;
if plotting
    figure
    hold on;
    plot(xExact, sigmaExact, '--k');
    plot([x(1:end-1)', x(2:end)']', [stresses(1:end), stresses(1:end)]', '-r');
    hold off;
end

uComputed = zeros(size(xExact));
for element=1:size(elements,1)
    xEval = (xExact>=x(elements(element,1))) & (xExact<=x(elements(element,2)));
    uComputed(xEval) = displacementVector(elements(element,1)) + ...
        (displacementVector(elements(element,2))-displacementVector(elements(element,1)))/elementLengths(element) * (xExact(xEval)-x(elements(element,1)));
end

erVal = abs(uComputed-uExact);
if plotting
    figure
    plot(xExact, erVal, '-b');
end

maxError = max(erVal);

end